/**********************************************************************************
 * NL nl.c
 * =================
 * versione    02
 * autore      Calzoni Pietro aka Calzo <calzog@gmail.com>
 *               LUGMan - Linux Users Group Mantova
 *               www.lugman.org
 * licenza     GPL
 * Kernel      2.6
 * Kernel test 2.6.13.1
 * System      SalckWare 10.2
 * compilers   gcc-3.3.6
 * --------------------------------------------------------------------------------
 * Dalla versione 01a si  visto come sia possibile ottenere un messaggio dallo
 * user-space, ma questo messaggio non  stato elaborato.
 * In questo modulo si vuole ora estrare i dati inviati dallo user-space e
 * stamparli nel ring buffer.
 * Siccome ora cominceranno a dover essere definite alcune macro che verranno usate
 * anche nello userspace.
 * --------------------------------------------------------------------------------
 * IMPEMENTAZIONE:
 * l'implementazione  relativa principalmente alla callback input() associata al
 * socket. L'accesso ai dati provenienti da user-space viene fatto con:
 * - sk_buff: definito in skbuff.h. Ha una serie infinita di campi anche a union
 *  e tra gli altri permette l'accesso diretto al buffer del socket tramite il
 *  campo data. Questo campo contiene anche l'header nlmsghdr.
 * - nlmsghdr: NETLINK Message Header definito in linux/netlink.h. Tra i vari campi
 *  continene anche nlmsg_pid ossia il pid del progesso mittente del messaggio.
 *********************************************************************************/
#include <linux/init.h>
#include <linux/module.h>
#include <linux/types.h>
//--- NetLink ---
#include <linux/socket.h>  // gestione dei socket e strutture relative
#include <linux/netlink.h> // implementato in net/netlink/af_netlink.c (2.6.13.1)
#include <net/sock.h>      // implementa la struttura sock. senza questo  un casino referenziare i puntatori
//--- nl.h ---
#include "nl.h"            // comprende macro e variabili necessarie al kernel quanto allo user space

MODULE_LICENSE("GPL");
MODULE_VERSION("0.2");

static struct sock *nl_sk = NULL;

/** _input  una callback che deve essere definita e passata alla funzione di
 * creazione del ntlink socket `netlink_kernel_create()'.
 * NOTA: se il manage del socket  abbastanza leggero, la funzione seguente pu
 * occuparsene direttamente, altrimenti  bene eseguire un wake_up_ in modo da
 * sbloccare l'esecuzione di un task sospeso appositamente.
 * */
void nl_input(struct sock *sk, int len)
{
   struct sk_buff *buf = NULL;
   struct nlmsghdr *nlh;
   int err;
   char *payload;
   
   printk(KERN_ALERT "Messaggio ricevuto da user-space\n");
   
   // Prelevo dalla coda il buffer. Per gestire tutta la coda dei messaggi credo sia mglio
   // usare skb_dequeue(&sk->receive_queue)
   buf = skb_recv_datagram(sk, 0, 0, &err);
   nlh = (struct nlmsghdr *)buf->data; // prelevo il la struttura del messaggio compreso l'header NETLINK
   payload = NLMSG_DATA(nlh);          // prelevo il payload, ossia i dati effettivi
   printk(KERN_ALERT " sender pid: %d > %s\n", 
	  nlh->nlmsg_pid, payload);
}


/** _init_module non fa altro che creare il socket NETLINK.
 * La creazione del socket a livello kernel si esegue con `netlink_kernel_create()'.
 * Non si usa netlink_create() la quale viene comunque richiamata da _kernel_create.
 * */
int nl_init_module(void)
{
   nl_sk = netlink_kernel_create(NETLINK_TEST,
				 nl_input);
   if(nl_sk)
     printk(KERN_INFO "Socket creato: %u\n", (unsigned)nl_sk);
   else
     printk(KERN_ALERT "Socket non creato: %u\n", (unsigned)nl_sk);
   
   return 0;
}


/** _cleanup_module deve assolutamente eliminare il socket NETLINK.
 * NOTA: cosa succede sa il socket  aperto (x es in user-space) e scarico il 
 * modulo?
 * */
void nl_cleanup_module(void)
{
   printk(KERN_INFO "Close NETLINK socket %u...\n", (unsigned)nl_sk);
   sock_release(nl_sk->sk_socket);
}

module_init(nl_init_module);
module_exit(nl_cleanup_module);
