/*********************************************************************************
 * nl_send: permette di spedire un messaggio al etlink del kernel creato con il
 * modulo nl.
 * Il socket NETLINK  descritto da un header di tipo sockaddr_nl accessibile
 * aprendo il socket in moalit RAW.
 * Per il driver versione 0.3 il programma  modificato. Ora posso inviare un
 * messaggio indicando quale pid  il mittente e a quale pid il kernel dovr
 * rispondere. Successicamente mi pongo in attesa di un messaggio.
 * Il PID pu essere falsificato passandolo come primo parametro ;)
 * -------------------------------------------------------------------------------
 * MODALITA' DI TEST:
 * Dopo aver caricato il modulo, si deve eseguire nl_send. Non  necessario essere 
 * root. Ad ogni modo nei miei test il programma appartiene all'utente e il grupo
 * di default (a cui appartiene anche il programma)  root, ma non dovrebbe essere
 * vincolante.
 * 
 * 1) Send e Receive dal processo stesso: si pu lanciare semplicemente nl_send in
 * modo che lui invii un messaggio di default al kernel e quest'ultimo reinvier il
 * messaggio al processo. Se si vuole cambiare il messaggio natenendo il PID,
 * lanciare:   nl_send -1 "novo messaggio"
 * 
 * 2) Send con pid falsificato: nl_send 12345 "nuovo messaggio" (1245  un pid a
 * caso; pu essere anche 0) In questo caso il kernel invier la risposta a 12345
 * e quindi il nostro processo rimarr in attesa. Lo si pu chiudere con Ctrl-C.
 * 
 * 3) Send con pid falsificato, ma noto: lanciare nell'ordine
 * $ nl_send 12345 "nuovo messaggio": questo prima di inviare il messaggio vi dice
 * qual' il suo PID, per esempio 8100 
 * $ nl_send 8100 "Altro Messaggio": questo programma invia il messaggio, ma il
 * kernel invier il messaggio al processo nl_send lanciato in precedenza il quale
 * ricever "aLTRO mESSAGGIO" e si chiuder, mentre il secondo nl_send rimarr in
 * attesa di un messaggio che forse non arriver mai... che triste storia
 * -------------------------------------------------------------------------------
 *  NOTA: una volta costruita la struttura msg (che comprende iov, nhl e payload)
 * questa funzione verr usata sempre sia per ricezione che per trasmissione. Ci
 *  anche logico da come viene costruita in quanto msg non viene creata con dei
 * memcpy, ma semplicemente indirizzando i vari campi come puntatori.
 *  NOTA: questo sistema di comunicazione  facilmente falsificabile anche senza
 * scomodare una struttura ad hoc come abbiamo fatto noi, ma basta in teoria
 * modificare a proprio piacimento il campo nlh->nlmsg_pid con un PID a noi noto
 * -------------------------------------------------------------------------------
 * BUG: con kernel 2.6.13.1 e compilatore 3.3.6 credo ci sia un bug perch
 * sys/socket.h deve andare prima di linux/netlink.h, ma non capisco perch visto
 * che le macro di esclusione sono diverse!! _
 *********************************************************************************/

#include <stdio.h>
#include <sys/socket.h>    // deve essere prima di netlink.h altrimenti non si compila nulla!  un baco credo ma non capisco perch visto che le macro di esclusione sono diverse
#include <linux/netlink.h>

#include "nl.h"

int sock_fd;
struct sockaddr_nl src_addr, dest_addr;
struct nlmsghdr *nlh, *nlh_recv; // header di spedizione e ricezione
struct msghdr msg; 
struct iovec iov;

struct nl_payload_t p;

int main(int argc, char **argv)
{  
   int altpid = -1;  // pid alterativo
   
   sock_fd = socket(PF_NETLINK,    // il socket  un Kernel User Interface Device
		    SOCK_RAW,      // acesso grezzo (RAW)
		    NETLINK_TEST); // 16  lo stesso numero definito nel kernel
   
   src_addr.nl_family = AF_NETLINK;
   src_addr.nl_pid    = getpid();
   src_addr.nl_groups = 0;         // gruppo a cui interessano i messaggi multicast. Se 0 non ci sono multicast
   
   if(bind(sock_fd, (struct sockaddr*)&src_addr,
	   sizeof(src_addr)))
     goto exit_;
   
   memset(&dest_addr, 0, sizeof(src_addr));
   dest_addr.nl_family= AF_NETLINK;
   dest_addr.nl_pid   = 0;         // messaggio destinato al kernel
   dest_addr.nl_groups= 0;         // unicast
   
   nlh = (struct nlmsghdr*)malloc(NLMSG_SPACE(MAX_PAYLOAD_LENGTH));
   nlh->nlmsg_len = NLMSG_SPACE(MAX_PAYLOAD_LENGTH);
   nlh->nlmsg_pid = getpid();  /* self pid, ma non serve */
   nlh->nlmsg_flags = 0;

   memset(&p, 0, sizeof(p));          // azzero il buffer
   
   if(argc > 2)     // se esiste il 2 allora  una stringa
     strncpy(p.str, argv[2], MAX_PAYLOAD_CHAR-1); // non copio pi del necessario. Il -1  per garantire la null-termination 
   else
     strncpy(p.str, "Calzo Message!!", MAX_PAYLOAD_CHAR-1);
   
   if(argc > 1)     // se esiste il 1, allora  un pid
     altpid = atoi(argv[1]);
   if(altpid < 0)   // se  minore di 0 usa il pid del processo...
     p.sender_pid = getpid();
   else             // ...altrimetni fissalo tu
     p.sender_pid = altpid;

   /* Fill in the netlink message payload */
   memcpy(NLMSG_DATA(nlh), &p, MAX_PAYLOAD_LENGTH);
   
   iov.iov_base = (void *)nlh;
   iov.iov_len = nlh->nlmsg_len;
   msg.msg_name = (void *)&dest_addr;
   msg.msg_namelen = sizeof(dest_addr);
   msg.msg_iov = &iov;
   msg.msg_iovlen = 1;

   printf("My PID: %d\tPress return to continue.", getpid());
   fflush(stdout);
   getchar();        // aspetto in modo da poter eseguire anche un altro processo
   printf("Proces %d send '%s' message...\n", nlh->nlmsg_pid, p.str);
   sendmsg(sock_fd, &msg, 0);
   
   /* A questo punto mi pongo in ricezione */
   recvmsg(sock_fd, &msg, 0);
   
   memcpy(&p, NLMSG_DATA(nlh), MAX_PAYLOAD_LENGTH);
   printf("Receiving message '%s' from %d program\n",
	  p.str, p.sender_pid);
   
exit_:
   close(sock_fd);
   return 0;
}

