/**********************************************************************************
 * NL nl.c
 * =================
 * versione    05
 * autore      Calzoni Pietro aka Calzo <calzog@gmail.com>
 *               LUGMan - Linux Users Group Mantova
 *               www.lugman.org
 * licenza     GPL
 * Kernel      2.6
 * Kernel test 2.6.13.1
 * System      SalckWare 10.2
 * compilers   gcc-3.3.6
 * --------------------------------------------------------------------------------
 * Non cambia nulla rispetto alla 04 dal punto di vista del funzionamento, ma solo
 * viene implementato il codice per kernel maggiori del 2.6.14. Si noti che in
 * questa serie di kernel NON c' il file linux/version.h e le #define relative
 * al numero del kernel vengono crete dal Makefile. Roba da prendere qualche
 * "sviluppatore" e prenderlo a calci nei coglioni. "Ovviamente" se il kenrel  !=
 * dal 14 tutti i file esistono... x fortuna.
 * Dal kernel 2.6.14 sono cambiate in netlink.h:
 * - struct netlink_skb_parms: ridefinisce alcuni parametri e ne aggiunge altri
 *   relativo a SELinux. Quelli cancellati erano quelli che avevamo detto essere
 *   inutili.
 * - sock* netlink_kernel_create: aggiunge un "gruppo", mutex e module come
 *   parametri. Attualmente metto tutto a NULL e ho lo stesso funzionamento di
 *   prima.
 * --------------------------------------------------------------------------------
 * IMPEMENTAZIONE:
 * l'implementazione  relativa principalmente alla callback input() associata al
 * socket. L'accesso ai dati provenienti da user-space viene fatto con:
 * - sk_buff: definito in skbuff.h. Ha una serie infinita di campi anche a union
 *  e tra gli altri permette l'accesso diretto al buffer del socket tramite il
 *  campo data. Questo campo contiene anche l'header nlmsghdr.
 * - nlmsghdr: NETLINK Message Header definito in linux/netlink.h. Tra i vari campi
 *  continene anche nlmsg_pid ossia il pid del progesso mittente del messaggio.
 * Occorre decidere come "pilotare" l'invio del messaggio da parte del kernel. Ci
 * lo si pu fare in molti modi (proc, device driver, ecc), ma la via pi semplice
 * forse  inviare un messaggio con la richiesta di reinvio del messaggio.
 * L'invio di un messaggio unicast diretto a uno specifico processo da parte del 
 * kernel viene effettuato con la chiamata:
 *            netlink_unicast(struct sock, struct sk_buff, u32, int);
 * L'invio di un messaggio multicast viene eseguito con la chiamata:
 *           netlink_broadcast(struct sock, struct sk_buff, u32, 
 *                             u32 group, int allocation);
 * Come si nota la strtuttura necessaria  sempre sk_buff a cui si accede mediate
 * la macro NETLINK_CB() definita in linux/netlink.h
 * Per quanto riguarda i messaggi broadcast, il gruppo  un OR dei gruppi a cui il
 * messaggio  destinato, mantre allocation  il tipo di allicazione che il kernel
 * dovr eseguire per clonare il messaggio.
 * E' stato poi introdotto il parametro `kernel_group' che permette di settare il
 * gruppo di appartenenza del kernel. Per fare questo baster modificare il
 * parametro tramite interfacci sysfs. si noti che questo gruppo non verr settato
 * fino a che il sistema non ricever un messaggio. La certezza del gruppo del
 * kernel  data da `actual_kernel_group'. Se i parametri sono uguali, il gruppo
 * sar settato correttamente.
 * --------------------------------------------------------------------------------
 * TEST:
 * Il test  condotto con nl_com per la versione 04 del modulo. Riferirisi alla
 * documentazione interna al file nl_mcom.c per vedere le varie modalit di test.
 * Anche il programma nl_com pu testare il modulo, ma  obsoleto.
 * --------------------------------------------------------------------------------
 * BUG:
 *  KERNEL: assertion (!atomic_read(&sk->sk_rmem_alloc)) failed at net/netlink/af_netlink.c (122)
 * Non capisco a cosa sia dovuto. Da quanto letto nei commenti relativi alla
 * `netlink_kernel_create()' sembra che il buffer del socket vada liberato sempre
 * prima di uscire, ma il tentativo di liberare la memoria ha portato anche a gravi
 * freeze del sistema.
 * Comunque il problema non sembra grave e pare sia generato solo dopo aver tolto
 * il driver. Forse  una race (o simile) dovuta al fatto che la liberazione del 
 * socket elimina la struttura dellle "operazioni" `netlink_ops'
 *********************************************************************************/
#include <linux/init.h>
#include <linux/module.h>
#include <linux/types.h>
//--- vrsione del kernel ---
#include <linux/version.h> // ATTENZIONE: nella versione 14 del kernel questo file non c'
//--- NetLink ---
#include <linux/socket.h>  // gestione dei socket e strutture relative
#include <linux/netlink.h> // implementato in net/netlink/af_netlink.c (2.6.13.1)
#include <net/sock.h>      // implementa la struttura sock. senza questo  un casino referenziare i puntatori
//--- nl.h ---
#include "nl.h"            // comprende macro e variabili necessarie al kernel quanto allo user space

MODULE_LICENSE("GPL");
MODULE_VERSION("0.4");
MODULE_AUTHOR("Calzo");
MODULE_DESCRIPTION("NL creates a NetLink socket and test the comunication between user and kernel space");

static unsigned kernel_group = 0;        // Gruppi di appartenenza in attesa di definizione
static unsigned actual_kernel_group = 0; // Gruppo attuale di appartenenza reale in questo momento
static unsigned use_skb_dequeue = 0;     // Usa skb_dequeue() al posto di skb_recv_datagram()

module_param(kernel_group, uint, S_IRUGO|S_IWUSR);    // see LDD3 cap 2, pag 36 e il file linux/stat.h
module_param(actual_kernel_group, uint, S_IRUGO);     // see LDD3 cap 2, pag 36 e il file linux/stat.h
module_param(use_skb_dequeue, uint, S_IRUSR|S_IWUSR); // see LDD3 cap 2, pag 36 e il file linux/stat.h

MODULE_PARM_DESC(kernel_group, "Kernel Group for multicast comunications (fielad .dest_groups) ready to the nest step");
MODULE_PARM_DESC(actual_kernel_group, "Actual Kernel Group for multicast comunications (fielad .dest_groups");
MODULE_PARM_DESC(use_skb_dequeue, "Use `skb_dequeue()' instead of `skb_recv_datagram()'");

static struct sock    *nl_sk  = NULL;

/** _change_str_case non fa altro che cambiare da minuscola a maiuscola e vice
 * versa la stringa passatagli.
 * */
void nl_change_str_case(char *str, int maxstr_len)
{
   int i;
   
   for(i=0; str[i] && i<maxstr_len; i++)
     {
	if(str[i]>='A' && str[i]<='Z')
	  str[i]+=0x20;
	else
	  if(str[i]>='a' && str[i]<='z')
	    str[i]-=0x20;
     }
}


/** _input  una callback che deve essere definita e passata alla funzione di
 * creazione del ntlink socket `netlink_kernel_create()'.
 * Questa funzione invia dal kernel allo user-space un messaggio a fronte di un
 * altro messaggio proveniente dallo user-space. Questo serve solo per triggherare
 * l'evento di invio da parte del kernel ed  una scelta arbitraria.
 * In funzione da quanto richiesto dallo user-space il messaggio pu essere 
 * unicast o multicast. Alcune note relativamente a queste due modalit:
 * - unicast: viene lanciata con MSG_DONTWAIT in modo che non sia interrttiva. La
 *   macro  definita in linux/socket.h. Consultarlo per le altre modalit.
 * - multicast: la funzione invia a un gruppo di processi il messaggio.
 * Viene poi azzerato nlh->nlmsg_pid = 0; questo permette al processo di capire
 * che il mittenete  il kernel e cos potremo vedere se due processi possono
 * usare NetLink in user-space dove questo valore sar diverso da 0.
 * 
 * Si noti poi che  stata aggiunta a funzione `kfree_skb()' alla fine. In realt
 * dai commenti letti in net/core/datagram.c realtivamente a `skb_recv_datagram()'
 * la funzione migliore da utilizzare per liberare la memoria sarebbe
 * `skb_free_datagram()'. Per questa non fa altro che chiamare `kfree_skb()', ma
 * chiedendo 2 parametri! per questo mi sembra inutile chiamare `skb_free_datagram'
 * almeno fino al kernel 2.6.23.8 (controllato).
 * Ad ogni modo le funzioni di liberazione della memoria non sono state messe
 * perch hanno causato sempre crash gravi del modulo o peggio freeze dell'intero
 * sistema se poste nella funzione di clean_up.
 * --------------------------------------------------------------------------------
 *  NOTA: se il manage del socket  abbastanza leggero, la funzione seguente pu
 * occuparsene direttamente, altrimenti  bene eseguire un wake_up_ in modo da
 * sbloccare l'esecuzione di un task sospeso appositamente.
 *  NOTA: MSG_DONTWAIT permette di rendere la chiamata non-interruttiva. Credo che
 * questo provochi la perdita del messaggio se nessun socket  in ascolto, ma non
 * ne sono sicuro.
 * */
void nl_input(struct sock *sk, int len)
{
   struct sk_buff *buf = NULL;
   struct nlmsghdr *nlh;
   int err;
   struct nl_payload_t *payload;
   
   printk(KERN_ALERT "Message received by user-space\n");
   
   // Prelevo dalla coda il buffer. Per gestire tutta la coda dei messaggi credo sia mglio
   // usare skb_dequeue(&sk->receive_queue). Implementata in net/core/datagram.c
   if(use_skb_dequeue == 0)
     buf = skb_recv_datagram(sk,    // socket NETLINK
			     0,     // Flags=0? chiama una volta skb_dequeue. L'unico flag accettato  MSG_PEEK 
			     0,     // tempo di blocking della funzione. 0=funzione non bloccante (see sock_rcvtimeo() implementation)
			     &err);
   else
     {
__redo_skb_dequeue:
	buf = skb_dequeue(&sk->sk_receive_queue);
	if(!buf)  // se il buffer  nullo esco
	  return; // e non devo liberare la memoria
     }
   
   nlh = (struct nlmsghdr *)buf->data;                // prelevo la struttura del messaggio compreso l'header NETLINK
   payload = (struct nl_payload_t *)NLMSG_DATA(nlh);  // prelevo il payload, ossia i dati effettivi
   printk(KERN_ALERT "  sender pid: %d > %s\n", 
	  nlh->nlmsg_pid, payload->str);
   
   // A questo punto conosco il mittente (o il presunto tale). 
   // Setto i parametri per l'invio. In realt non servono a
   // nulla, se non il .pid che indica la provenienza. La
   // destinazione  definita dalle funzioni.
   // Ad ogni modo scrivo il codice fedele alla documentazione
   // di Linux Journal 7356.
#if LINUX_VERSION_CODE < 132622           // Se il kernel  < del 2.6.14...   
   NETLINK_CB(buf).groups = kernel_group; // grupo di appartenenza
   NETLINK_CB(buf).pid = 0;               // Il processo mittente  il kernel
   NETLINK_CB(buf).dst_pid = payload->pid;          // PID del destinatario per ora lo imposto a 0 ossia kernel, ma non dovrebbero esserci problemi
   NETLINK_CB(buf).dst_group = payload->multicast; // Multicast  communication: maschera proveniente da user-space
   actual_kernel_group = NETLINK_CB(buf).groups; //NETLINK_CB(buf).groups;    // Setto il vero gruppo del kernel notificando l'eventuale cambiameto (lo si legge da sysfs)
#else
   NETLINK_CB(buf).pid = 0;               // Il processo mittente  il kernel
   NETLINK_CB(buf).dst_group = payload->multicast; // Multicast  communication: mas$
   actual_kernel_group = -1;              // -1, perch il kernel  >2.6.13
#endif   
   // Reinvio un messaggio unicast al pid
   if(!payload->multicast)
     if(payload->pid == 0)  // Se viene richiesto di inviare il messaggio AL
       {                    //  kernel, ritorno per evitare loop  ;)
	  printk(KERN_ALERT "Message sent by kernel: avoid processing\n");
	  goto __return;    // se non ritorno, il sistema va in freeze... ma di quelli potenti
       }

   // Rielaboro il Payload tanto per fare vedere qualche cosa
   nl_change_str_case(payload->str, MAX_PAYLOAD_CHAR);
   nlh->nlmsg_pid = 0;
   
   if(payload->multicast)  /// controllo se mi  richiesto di spedire in Multicast...
     {
	payload->pid = 0;                      // lo azzero per precauzione...  non si sa mai col multicast
	netlink_broadcast(nl_sk, buf, 
			  payload->pid,        // PID di destinazione indicato dallo user-space
			  payload->multicast,  // Multicast di destinazione indicato dallo user-space
			  GFP_KERNEL);         // Modalit di allocazione (GFP_ATOMIC sarebbe pi efficiente)
	printk(KERN_ALERT "  MULTICAST Msg '%s' sent with group mask 0x%X\n",
	       payload->str, payload->multicast);
     } else                /// ...ma se sono qui, spedisco in Unicast
     {
	netlink_unicast(nl_sk, buf, 
			payload->pid,   // PID di destinazione indicato dallo user-space  
			MSG_DONTWAIT);  // Chiamata non interruttiva
	printk(KERN_ALERT "  UNICAST Msg '%s' resent to %d process\n", 
	       payload->str,
	       payload->pid); // NETLINK_CB(buf).dst_pid);
     }
   
__return:
   if(use_skb_dequeue)
     goto __redo_skb_dequeue;
}


/** _init_module non fa altro che creare il socket NETLINK.
 * La creazione del socket a livello kernel si esegue con `netlink_kernel_create()'.
 * Non si usa netlink_create() la quale viene comunque richiamata da _kernel_create.
 * */
int nl_init_module(void)
{
#if LINUX_VERSION_CODE < 132622      // Se il kernel  < del 2.6.14...
   nl_sk = netlink_kernel_create(NETLINK_TEST, // TEST: valore arbitrario deciso da noi
				 nl_input);    // Funzione di gestione del socket
#else      // se siamo nel kernel 2.6.14 o +...
   nl_sk = netlink_kernel_create(NETLINK_TEST, // TEST: valore arbitrario deciso da noi
				 NETLINK_TEST,
				 nl_input,     // Funzione di gestione del socket
				 NULL, THIS_MODULE);
#endif
   if(nl_sk)
     printk(KERN_INFO "Socket created: %u\n", (unsigned)nl_sk);
   else
     printk(KERN_INFO "Socket not created: %u\n", (unsigned)nl_sk);

   return 0;
}


/** _cleanup_module deve assolutamente eliminare il socket NETLINK.
 * NOTA: cosa succede se il socket  aperto (x es in user-space) e scarico il 
 * modulo? ... uno strano messaggio
 * */
void nl_cleanup_module(void)
{
   printk(KERN_INFO "Close NETLINK socket %u...\n", (unsigned)nl_sk);
   sock_release(nl_sk->sk_socket);
}

module_init(nl_init_module);
module_exit(nl_cleanup_module);
