/**********************************************************************************
 * NL_COM nl_com.c 
 * =================
 * versione    01
 * autore      Calzoni Pietro aka Calzo (LUGMan - Linux Users Group Mantova)
 * licenza     GPL
 * Kernel      2.6
 * Kernel test 2.6.13.1
 * System      SalckWare 10.2
 * compilers   gcc-3.3.6
 * -------------------------------------------------------------------------------
 * Permette di comunicare con il socket NetLink del kernel appositamente creato
 * tramite il protocollo TEST. Il programma deve potersi porre in 2 modalit:
 * ricezione e trasmissione.
 * - ricezione: il sistema si pone in ascolto perenne. Termia con un Ctrl-C o
 *   inviando come primo carattere uno '0'
 * - trasmissione: il programma chiede come settare il messaggio da spedire
 * Il sistema non prevede l'uso di fork() per evitare di complicare inutilmente
 * il programma il quale ha solo una valenza didattica.
 * -------------------------------------------------------------------------------
 * MODALITA' DI TEST:
 * L'ascolto in multicast deve essere fatto come root
 * TODO
 * -------------------------------------------------------------------------------
 *  NOTA: una volta costruita la struttura msg (che comprende iov, nhl e payload)
 * questa funzione verr usata sempre sia per ricezione che per trasmissione. Ci
 *  anche logico da come viene costruita in quanto msg non viene creata con dei
 * memcpy, ma semplicemente indirizzando i vari campi come puntatori.
 *  NOTA: questo sistema di comunicazione  facilmente falsificabile anche senza
 * scomodare una struttura ad hoc come abbiamo fatto noi, ma basta in teoria
 * modificare a proprio piacimento il campo nlh->nlmsg_pid con un PID a noi noto
 * -------------------------------------------------------------------------------
 * BUG: con kernel 2.6.13.1 e compilatore 3.3.6 credo ci sia un bug perch
 * sys/socket.h deve andare prima di linux/netlink.h, ma non capisco perch visto
 * che le macro di esclusione sono diverse!! _
 *********************************************************************************/

#include <stdio.h>
#include <errno.h>
#include <sys/socket.h>    // deve essere prima di netlink.h altrimenti non si compila nulla!  un baco credo ma non capisco perch visto che le macro di esclusione sono diverse
#include <linux/netlink.h>

#include "nl.h"

#define QUIT_CHAR  '0'
#define str_l(x) "%"#x"s"
#define str_lim(x) str_l(x)

int    sock_fd;
struct sockaddr_nl src_addr, dest_addr;
struct nlmsghdr *nlh, *nlh_recv; // header di spedizione e ricezione
struct msghdr msg; 
struct iovec iov;

struct nl_payload_t p;

void   nl_receiver(void);      // funzione di ricezione
void   nl_transmitter(void);   // funzione di trasmissione
int    nl_bind(int);           // funzione di bind con settaggio di src_addr 

int main(int argc, char **argv)
{  
   int retval=0;
   
   sock_fd = socket(PF_NETLINK,    // il socket  un Kernel User Interface Device
		    SOCK_RAW,      // acesso grezzo (RAW)
		    NETLINK_TEST); // 16  lo stesso numero definito nel kernel
   
   memset(&p, 0, sizeof(p));       // azzero per precauzione il payload
   memset(&dest_addr, 0, sizeof(src_addr));
   dest_addr.nl_family= AF_NETLINK;
   dest_addr.nl_pid   = 0;         // messaggio destinato al kernel
   dest_addr.nl_groups= 0;         // unicast
   
   nlh = (struct nlmsghdr*)malloc(NLMSG_SPACE(MAX_PAYLOAD_LENGTH));
   nlh->nlmsg_len = NLMSG_SPACE(MAX_PAYLOAD_LENGTH);
   nlh->nlmsg_pid = getpid();  /* self pid, ma non serve */
   nlh->nlmsg_flags = 0;

   iov.iov_base = (void *)nlh;
   iov.iov_len = nlh->nlmsg_len;
   msg.msg_name = (void *)&dest_addr;
   msg.msg_namelen = sizeof(dest_addr);
   msg.msg_iov = &iov;
   msg.msg_iovlen = 1;
   
   if(argc==1)     // se passo dei parametri sono in rtasmissione...
     {
	if(retval=nl_bind(0))  // bind con gruppo 0 se sono in trasmissione, ma un valore vale l'altro in questo caso specifico
	  goto exit_;
	nl_transmitter();
     }
   else            // ...altrimenti in ricezione
     {
	int bgroup;     // devo determinare il broadcast group
	
	bgroup = atoi(argv[1]); // il broadcast  passato come primo parametro
	if(bgroup<0)
	  bgroup = 0;
	
	if(bgroup)      // controllo se il gruppo  veramente multicast (!=0)
	  if (getuid() && geteuid()) // se sono qui sono in multicast e DEVO essere root
	    {
	       fprintf(stderr, "Multicast Bind must be run as root!\n");
	       retval=-EPERM; // operazione non permessa
	       goto exit_;
	    }

	if(retval=nl_bind(bgroup))  // bind con gruppo 0 se sono in trasmissione, ma un valore $
	  goto exit_;
	nl_receiver();
     }
   
exit_:
   printf("Quit process %d...\n", getpid());
   if(nlh)
     free(nlh);
   close(sock_fd);
   
   return retval;
}



/** nl_receiver: funzione di ricezione
 * */
void nl_receiver(void)
{
   printf("Receiver PID: %d\n", getpid());
   do {
      recvmsg(sock_fd, &msg, 0);
      
      memcpy(&p, NLMSG_DATA(nlh), MAX_PAYLOAD_LENGTH);
      printf("PID %d:\tReceiving message '%s' from %d process\n",
	     getpid(), p.str,  nlh->nlmsg_pid);
   }
   while(p.str[0]!=QUIT_CHAR);
}



/** nl_transmitter: funzione di trasmissione che chiede di compilare i vari campi
 * */
void nl_transmitter(void)
{
   printf("Transmitter PID: %d\nMessages will be sent to kernel.\n\n", getpid());
   do {
      
      memset(&p, 0, sizeof(p));          // azzero il buffer
      
      printf("Message:\t");
      fflush(stdout);
      //scanf(str_limit(MAX_PAYLOAD_CHAR), p.str);
      scanf(str_lim(MAX_PAYLOAD_CHAR), p.str);
      p.str[MAX_PAYLOAD_CHAR-1]='\0'; // per sicurezza termino la stringa
      
      printf("Multicast mask:\t");
      fflush(stdout);
      scanf("%d", &p.multicast);
      
      printf("PID:\t");
      fflush(stdout);
      scanf("%d", &p.pid);
      
      /* Fill in the netlink message payload */
      memcpy(NLMSG_DATA(nlh), &p, MAX_PAYLOAD_LENGTH);
      
      printf("Process %d send '%s' message...\n\n", nlh->nlmsg_pid, p.str);
      sendmsg(sock_fd, &msg, 0);
 
   } while(p.str[0]!=QUIT_CHAR); 
}



/** nl_bind: esegue in binding dal socket NETLINK settando il gruppo di appartenenza
 * per i messaggi broadcast. Se 0 il gruppo ricever solo messaggi unicast ossia
 * indirizzati tramite PID
 * */
int nl_bind(int multicast_group)
{
   printf("Multicast Group: %d\n", multicast_group);
   
   src_addr.nl_family = AF_NETLINK;
   src_addr.nl_pid    = getpid();
   src_addr.nl_groups = multicast_group; // gruppo a cui interessano i messaggi multicast$
   
   return bind(sock_fd, (struct sockaddr*)&src_addr,
	       sizeof(src_addr));
}
