/**********************************************************************************
 * NL_MCOM nl_mcom.c 
 * =================
 * versione    01
 * autore      Calzoni Pietro aka Calzo <calzog@gmail.com>
 *               LUGMan - Linux Users Group Mantova
 *               www.lugman.org
 * licenza     GPL
 * Kernel      2.6
 * Kernel test 2.6.13.1
 * System      SalckWare 10.2
 * compilers   gcc-3.3.6
 * -------------------------------------------------------------------------------
 * E' praticamente la versione 2 di nl_com con la differenza che ora  possibile
 * definire il tipo di comunicazione multicast da user-space, mentre l'uicast pu
 * essere indirizzato a un PID diverso dal kernel.
 * 
 * Permette di comunicare con il socket NetLink del kernel appositamente creato
 * tramite il protocollo TEST. Il programma deve potersi porre in 2 modalit:
 * ricezione e trasmissione.
 * - ricezione: il sistema si pone in ascolto perenne. Termia con un Ctrl-C o
 *   inviando come primo carattere uno '0'
 * - trasmissione: il programma chiede come settare il messaggio da spedire, prima
 *   chiedendo a quale processo operando sui campi .nl_pid e .nl_groups. Se il 
 *   processo  kernel (0) verr chiesto di completare il resto dei campi della 
 *   struttura nl_payload_t oltre a message. Questo perch, si ricorda, i campi
 *   serviranno al kernel per creare il suo messaggio.
 * Il sistema non prevede l'uso di fork() per evitare di complicare inutilmente
 * il programma il quale ha solo una valenza didattica.
 * -------------------------------------------------------------------------------
 * MODALITA' DI TEST:
 * L'ascolto in multicast deve essere fatto come root. 
 * Didascalia: US=user-space, KS=kernel-space, #=numero casuale (normalmente non 0
 * ne un PID noto), S= programma trasmittente, R1,R2,Rn=pid programmi riceventi, 
 * G1,G2,Gn=valore del gruppo dei programmi riceventi 1,2,n
 * Tra i test principali si segnala: 
 * (US=user-space, KS=kernel-space, #=numero casuale)
 * - Send da US a KS unicast msg, impostare: dest_pid=0, dest_group=0, msg=xyz,
 *   pid=#, multicast=#. Si verifica solo la trasmissione al kenel.
 * - Send da US a US unicast msg, impostare: dest_pid=R1, dest_group=0, msg=xyz.
 *   Si verifica che il processo in ascolto riceva xyz
 * - Send da US a US multicast msg, impostare: dest_pid=0, dest_group=G1|G2,
 *   msg=xyz. Si verifica che il programma R1 e R2 ricevono il messaggio,ma non R3
 * - Send da US a R1 tramite gruppo e a R2 tramite PID: dest_pid=R1, dest_group=G2
 *   msg=xyz. Si verifica che R1 e R2 avranno il messaggio, ma non R3
 * 
 * Per testare il comportamento del kernel, questo va istruito dal programma S:
 * - Send KS a US unicast:
 * TODO TODO TODO TODO TODO TODO 
 * TODO TODO TODO TODO TODO TODO 
 * -------------------------------------------------------------------------------
 *  NOTA: una volta costruita la struttura msg (che comprende iov, nhl e payload)
 * questa funzione verr usata sempre sia per ricezione che per trasmissione. Ci
 *  anche logico da come viene costruita in quanto msg non viene creata con dei
 * memcpy, ma semplicemente indirizzando i vari campi come puntatori.
 *  NOTA: questo sistema di comunicazione  facilmente falsificabile anche senza
 * scomodare una struttura ad hoc come abbiamo fatto noi, ma basta in teoria
 * modificare a proprio piacimento il campo nlh->nlmsg_pid con un PID a noi noto
 * -------------------------------------------------------------------------------
 * BUG: con kernel 2.6.13.1 e compilatore 3.3.6 credo ci sia un bug perch
 * sys/socket.h deve andare prima di linux/netlink.h, ma non capisco perch visto
 * che le macro di esclusione sono diverse!! _
 *********************************************************************************/

#include <stdio.h>
#include <errno.h>
#include <sys/socket.h>    // deve essere prima di netlink.h altrimenti non si compila nulla!  un baco credo ma non capisco perch visto che le macro di esclusione sono diverse
#include <linux/netlink.h>

#include "nl.h"

#define QUIT_CHAR  '0'

int    sock_fd;
struct sockaddr_nl src_addr, dest_addr; // dest_ non ha influenza se siamo in ricezione
struct nlmsghdr  *nlh;  // header di spedizione e ricezione
struct msghdr    msg;   // richiesto per la funzione `sendmsg()'
struct iovec     iov;   // serve a msg

struct nl_payload_t p;  // formattazione arbitraria del payload

void   nl_config_dest(unsigned, unsigned); // funzione di configurazione del socket per il target d destinazione
void   nl_receiver(void);      // funzione di ricezione
void   nl_sender(void);        // funzione di trasmissione
int    nl_bind(int);           // funzione di bind con settaggio di src_addr 

int main(int argc, char **argv)
{  
   int retval=0;
   
   sock_fd = socket(PF_NETLINK,    // il socket  un Kernel User Interface Device
		    SOCK_RAW,      // acesso grezzo (RAW)
		    NETLINK_TEST); //  lo stesso numero definito nel kernel
   
   memset(&p, 0, sizeof(p));       // azzero per precauzione il payload
   memset(&dest_addr, 0, sizeof(src_addr));

   nl_config_dest(0,0);
   
   if(argc==1)     // se passo dei parametri sono in trasmissione...
     {
	if(retval=nl_bind(0))  // bind con gruppo 0 se sono in trasmissione, ma un valore vale l'altro in questo caso specifico
	  goto exit_;
	nl_sender();
     }
   else            // ...altrimenti in ricezione
     {
	int bgroup;     // devo determinare il broadcast group
	
	bgroup = atoi(argv[1]); // il broadcast  passato come primo parametro
	if(bgroup<0)
	  bgroup = 0;
	
	if(bgroup)      // controllo se il gruppo  veramente multicast (!=0)
	  if (getuid() && geteuid()) // se sono qui sono in multicast e DEVO essere root
	    {
	       fprintf(stderr, "Multicast Bind must be run as root!\n");
	       retval=-EPERM; // operazione non permessa
	       goto exit_;
	    }

	if(retval=nl_bind(bgroup))  // bind con gruppo 0 se sono in trasmissione, ma un valore $
	  goto exit_;
	nl_receiver();
     }
   
exit_:
   printf("Quit process %d...\n", getpid());
   if(nlh)
     free(nlh);
   close(sock_fd);
   
   return retval;
}

/** nl_config_dest: funzione di configurazione del messaggio. Serve soprattutto per
 * la trasmissione. Le variabili "libere" di essere configurate sono il PID del
 * processo (0 per il kernel), mentre il gruppo indica i gruppi a cui il messaggio
 *  destinato.
 * */
void nl_config_dest(unsigned dest_pid, unsigned dest_group)
{
   dest_addr.nl_family = AF_NETLINK;
   dest_addr.nl_pid    = dest_pid;   // messaggio destinato al kernel
   dest_addr.nl_groups = dest_group; // unicast=0, multicat altrimenti
   
   nlh = (struct nlmsghdr*)malloc(NLMSG_SPACE(MAX_PAYLOAD_LENGTH));
   nlh->nlmsg_len = NLMSG_SPACE(MAX_PAYLOAD_LENGTH); // forse  NLMSG_LENGTH...??? VERIFICA. per verificarlo usa NLMSG_PAYLOAD
   nlh->nlmsg_pid = getpid();  // self pid, ma non serve molto per il funzionamento
   nlh->nlmsg_flags = 0;
   
   iov.iov_base = (void *)nlh;
   iov.iov_len = nlh->nlmsg_len;
   msg.msg_name = (void *)&dest_addr;
   msg.msg_namelen = sizeof(dest_addr);
   msg.msg_iov = &iov;
   msg.msg_iovlen = 1;
}


/** nl_receiver: funzione di ricezione
 * */
void nl_receiver(void)
{
   printf("Receiver PID: %d\n", getpid());
   do {
      recvmsg(sock_fd, &msg, 0);
      
      memcpy(&p, NLMSG_DATA(nlh), MAX_PAYLOAD_LENGTH);
      printf("PID %d:\tReceiving message '%s' from %d process\n",
	     getpid(), p.str,  nlh->nlmsg_pid);
   }
   while(p.str[0]!=QUIT_CHAR);
}



/** nl_sender: funzione di trasmissione che chiede di compilare i vari campi
 * */
void nl_sender(void)
{
   int c;
   unsigned dest_pid, dest_groups;
   
   dest_pid    = dest_addr.nl_pid;
   dest_groups = dest_addr.nl_groups;
   
   printf("Sender PID: %d\n\n", getpid());

   do {
	
      printf("----------------------------------\n"
	     "Destination message configuration:\n"
	     "  Destination PID:   %d\n"
	     "  Destination Group: 0x%X\n",
	     dest_addr.nl_pid, dest_addr.nl_groups);

      printf(" > change destination PID:\t");
      fflush(stdout);
      scanf("%d", &dest_pid);
      
      printf(" > change destination Group:\t");
      fflush(stdout);
      scanf("%d", &dest_groups);
      
      nl_config_dest(dest_pid, dest_groups);
      printf("New destination message configuration:\n"
	     "  Destination PID:   %d\n"
	     "  Destination Group: 0x%X\n",
	     dest_addr.nl_pid, dest_addr.nl_groups);
      
      printf("Fill payload structure for %s:\n", dest_pid ? "process" : "kernel");
      memset(&p, 0, sizeof(p));        // azzero il buffer
      
      printf("Message:\t");
      fflush(stdout);
      scanf(str_lim(MAX_PAYLOAD_CHAR), p.str);
      p.str[MAX_PAYLOAD_CHAR-1]='\0'; // per sicurezza termino la stringa
      
      if(!dest_pid)     // se la destinazione  il kernel, ossia se dest_pid==0...
	{
	   printf("Multicast mask:\t");
	   fflush(stdout);
	   scanf("%d", &p.multicast);
	   
	   printf("PID:\t");
	   fflush(stdout);
	   scanf("%d", &p.pid);
	}
            
      /* Fill in the netlink message payload */
      memcpy(NLMSG_DATA(nlh), &p, MAX_PAYLOAD_LENGTH);
      
      printf("Process %d send '%s' message...\n\n", nlh->nlmsg_pid, p.str);
      sendmsg(sock_fd, &msg, 0);
 
   } while(p.str[0]!=QUIT_CHAR); 
}



/** nl_bind: esegue in binding dal socket NETLINK settando il gruppo di appartenenza
 * per i messaggi broadcast. Se 0 il gruppo ricever solo messaggi unicast ossia
 * indirizzati tramite PID
 * */
int nl_bind(int multicast_group)
{
   printf("Multicast Group: %d\n", multicast_group);
   
   src_addr.nl_family = AF_NETLINK;
   src_addr.nl_pid    = getpid();
   src_addr.nl_groups = multicast_group; // gruppo a cui interessano i messaggi multicast$
   
   return bind(sock_fd, (struct sockaddr*)&src_addr,
	       sizeof(src_addr));
}
